"use strict";

import React, { createContext, useCallback, useContext, useEffect, useRef } from 'react';
import { addCallback, addClean, consumeCallbacks, consumeCleans, removeCallback, removeClean, useStore } from "./store.js";
import { isFunction, parallelize } from "./utils.js";
import { jsx as _jsx } from "react/jsx-runtime";
const ConfigurationContext = /*#__PURE__*/createContext({
  enabled: true
});
const CallbacksContext = /*#__PURE__*/createContext({
  addInViewPortCallback: () => {},
  removeInViewPortCallback: () => {},
  addIsFirstCallback: () => {},
  removeIsFirstCallback: () => {}
});
const IsInViewPortContext = /*#__PURE__*/createContext({
  isInViewPort: () => undefined
});
const ItemContext = /*#__PURE__*/createContext({
  key: undefined
});
export const useInViewPort = (callback, deps) => {
  const {
    key
  } = useContext(ItemContext);
  const {
    addInViewPortCallback,
    removeInViewPortCallback
  } = useContext(CallbacksContext);
  const finalDeps = Array.isArray(deps) ? deps : [];
  useEffect(() => {
    if (!key) return; // If it is not an item of observable list.
    addInViewPortCallback(key, callback);
    return () => {
      removeInViewPortCallback(key, callback);
    };
    // The callback depends on deps.
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [key, addInViewPortCallback, removeInViewPortCallback, ...finalDeps]);
};
export const useIsFirst = (callback, deps) => {
  const {
    key
  } = useContext(ItemContext);
  const {
    addIsFirstCallback,
    removeIsFirstCallback
  } = useContext(CallbacksContext);
  const finalDeps = Array.isArray(deps) ? deps : [];
  useEffect(() => {
    if (!key) return; // If it is not an item of observable list.
    addIsFirstCallback(key, callback);
    return () => {
      removeIsFirstCallback(key, callback);
    };
    // The callback depends on deps.
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [key, addIsFirstCallback, removeIsFirstCallback, ...finalDeps]);
};
export function observe(List) {
  return /*#__PURE__*/React.forwardRef(function ({
    onViewableItemsChanged,
    keyExtractor,
    renderItem,
    $$enabled,
    ...props
  }, ref) {
    const {
      key
    } = useContext(ItemContext);
    const {
      isInViewPort
    } = useContext(IsInViewPortContext);
    const firstKey = useRef();
    const viewableKeys = useRef(new Set()).current;

    // traverse to root recursively.
    const isInViewPortRecursively = useCallback(itemKey => {
      const isInParentViewPort = isInViewPort(key);
      const inViewPort = viewableKeys.has(itemKey);
      if (isInParentViewPort === undefined) {
        return inViewPort;
      }
      return inViewPort && isInParentViewPort;
    }, [isInViewPort, key, viewableKeys]);
    const inViewPortStore = useStore();
    const isFirstStore = useStore();
    const addCallbackOrClean = useCallback((store, itemKey, callback) => {
      if (isInViewPortRecursively(itemKey)) {
        const clean = callback();
        addClean(store, itemKey, callback, clean);
      } else {
        addCallback(store, itemKey, callback);
      }
    }, [isInViewPortRecursively]);
    const removeCallbackTasks = useRef();
    const reserveRemoveCallbackAndClean = useCallback((store, itemKey, callback) => {
      const task = () => {
        removeCallback(store, itemKey, callback);
        removeClean(store, itemKey, callback);
      };
      if (!removeCallbackTasks.current) {
        removeCallbackTasks.current = [];
      }
      removeCallbackTasks.current.push(task);
    }, []);
    const {
      enabled: parentEnabled
    } = useContext(ConfigurationContext);
    const enabled = parentEnabled && $$enabled !== false ? true : false;
    const enabledRef = useRef(enabled); //  This is because `FlashList` does not detect changes in `onViewableItemsChanged`.
    enabledRef.current = enabled;

    // When self is an item of an observable list.
    useInViewPort(() => {
      if (!enabled) return;
      viewableKeys.forEach(itemKey => {
        if (isInViewPortRecursively(itemKey)) {
          consumeCallbacks(itemKey, inViewPortStore);
        }
      });
      return () => {
        viewableKeys.forEach(itemKey => {
          consumeCleans(itemKey, inViewPortStore);
        });
      };
    }, [enabled]);
    useEffect(() => {
      // true -> false
      if (!enabled) {
        viewableKeys.forEach(itemKey => {
          consumeCleans(itemKey, inViewPortStore);
        });
      }
      return () => {
        // false -> true
        if (!enabled) {
          // consume callback
          viewableKeys.forEach(itemKey => {
            if (isInViewPortRecursively(itemKey)) {
              consumeCallbacks(itemKey, inViewPortStore);
            }
          });
        }
      };
      // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [enabled]);
    const handleViewableKeys = useCallback(({
      viewableItems
    }) => {
      const willHideKeys = new Set(viewableKeys);
      for (let i = 0; i < viewableItems.length; i++) {
        const viewableItem = viewableItems[i];
        if (!viewableItem) {
          continue;
        }
        const {
          item,
          isViewable
        } = viewableItem;
        if (isViewable) {
          const itemKey = isFunction(keyExtractor) ? keyExtractor(item) : item;
          const isNew = !viewableKeys.has(itemKey);
          if (isNew) {
            viewableKeys.add(itemKey);
            if (enabledRef.current) {
              if (isInViewPortRecursively(itemKey)) {
                consumeCallbacks(itemKey, inViewPortStore);
              }
            }
          }
          willHideKeys.delete(itemKey);
        }
      }
      willHideKeys.forEach(itemKey => {
        viewableKeys.delete(itemKey);
        if (enabledRef.current) {
          consumeCleans(itemKey, inViewPortStore);
        }
      });
    }, [inViewPortStore, isInViewPortRecursively, keyExtractor, viewableKeys]);
    const handleFirstKey = useCallback(({
      viewableItems
    }) => {
      const first = viewableItems[0]?.item;
      const firstItemKey = isFunction(keyExtractor) ? keyExtractor(first) : first;
      if (enabledRef.current && firstItemKey !== firstKey.current) {
        const prevKey = firstKey.current;
        if (prevKey !== undefined) {
          consumeCleans(prevKey, isFirstStore);
        }
        firstKey.current = firstItemKey;
        if (isInViewPortRecursively(firstItemKey)) {
          consumeCallbacks(firstItemKey, isFirstStore);
        }
      }
    }, [isFirstStore, isInViewPortRecursively, keyExtractor]);
    const handleRemoveCallbackTasks = useCallback(() => {
      if (removeCallbackTasks.current) {
        for (let i = 0; i < removeCallbackTasks.current.length; i++) {
          removeCallbackTasks.current[i]?.();
        }
      }
      removeCallbackTasks.current = undefined;
    }, []);

    // eslint-disable-next-line react-hooks/exhaustive-deps
    const handleViewableItemsChanged = useCallback(parallelize(handleRemoveCallbackTasks, handleViewableKeys, handleFirstKey, onViewableItemsChanged), [handleRemoveCallbackTasks, handleViewableKeys, handleFirstKey, onViewableItemsChanged]);
    const ListComponent = List;
    return /*#__PURE__*/_jsx(ConfigurationContext.Provider, {
      value: {
        enabled
      },
      children: /*#__PURE__*/_jsx(IsInViewPortContext.Provider, {
        value: {
          isInViewPort: isInViewPortRecursively
        },
        children: /*#__PURE__*/_jsx(CallbacksContext.Provider, {
          value: {
            addInViewPortCallback: useCallback((itemKey, callback) => addCallbackOrClean(inViewPortStore, itemKey, callback), [addCallbackOrClean, inViewPortStore]),
            removeInViewPortCallback: useCallback((itemKey, callback) => reserveRemoveCallbackAndClean(inViewPortStore, itemKey, callback), [reserveRemoveCallbackAndClean, inViewPortStore]),
            addIsFirstCallback: useCallback((itemKey, callback) => addCallbackOrClean(isFirstStore, itemKey, callback), [addCallbackOrClean, isFirstStore]),
            removeIsFirstCallback: useCallback((itemKey, callback) => reserveRemoveCallbackAndClean(isFirstStore, itemKey, callback), [reserveRemoveCallbackAndClean, isFirstStore])
          },
          children: /*#__PURE__*/_jsx(ListComponent, {
            ...props,
            ref: ref,
            keyExtractor: keyExtractor,
            onViewableItemsChanged: handleViewableItemsChanged,
            renderItem: useCallback(itemProps => {
              const itemKey = isFunction(keyExtractor) ? keyExtractor(itemProps.item) : itemProps.item;
              return /*#__PURE__*/_jsx(ItemContext.Provider, {
                value: {
                  key: itemKey
                },
                children: renderItem(itemProps)
              });
            }, [keyExtractor, renderItem])
          })
        })
      })
    });
  });
}
//# sourceMappingURL=observe.js.map